from delfick_project.option_merge import MergedOptions
from delfick_project.norms import dictobj, Meta
import asyncio


class Command(dictobj.Spec):
    _merged_options_formattable = True

    async def execute(self):
        raise NotImplementedError("Base command has no execute implementation")


class Commander:
    """
    Entry point for creating an executor to execute commands with
    """

    _merged_options_formattable = True

    def __init__(self, store, **options):
        self.store = store

        everything = MergedOptions.using(options, {"commander": self}, dont_prefix=[dictobj])

        self.meta = Meta(everything, [])

    def process_reply(self, msg, exc_info):
        """Hook for every reply and progress message sent to the client"""

    def peek_valid_request(self, meta, command, path, body):
        """Hook for looking at every request"""

    def executor(self, progress_cb, request_handler, **extra_options):
        return Executor(self, progress_cb, request_handler, extra_options)


class Executor:
    _merged_options_formattable = True

    def __init__(self, commander, progress_cb, request_handler, extra_options):
        self.commander = commander
        self.progress_cb = progress_cb
        self.extra_options = extra_options
        self.request_handler = request_handler

    async def execute(
        self, path, body, extra_options=None, allow_ws_only=False, request_future=None
    ):
        """
        Responsible for creating a command and calling execute on it.

        If command is not already a Command instance then we normalise it
        into one.

        We have available on the meta object:

        __init__ options
            Anything that is provided to the Commander and Executor at __init__

        store
            The store of commands

        path
            The path that was passed in

        executor
            This executor

        request_future
            A future that is cancelled after execute is finished. This is the
            one passed in or a new Future for this request.

            At the end of the execute, the request future is cancelled, unless it
            was provided, in which case it is left alone.

        extra options
            Anything provided as extra_options to this function
        """
        provided = request_future is not None
        request_future = request_future or asyncio.Future()
        request_future._merged_options_formattable = True

        try:
            everything = MergedOptions.using(
                self.commander.meta.everything,
                {
                    "path": path,
                    "store": self.commander.store,
                    "executor": self,
                    "progress_cb": self.progress_cb,
                    "allow_ws_only": allow_ws_only,
                    "request_future": request_future,
                    "request_handler": self.request_handler,
                },
                self.extra_options,
                extra_options or {},
                dont_prefix=[dictobj],
            )

            meta = Meta(everything, self.commander.meta.path).at("<input>")
            execute = self.commander.store.command_spec.normalise(
                meta, {"path": path, "body": body, "allow_ws_only": allow_ws_only}
            )

            self.commander.peek_valid_request(meta, execute.__whirlwind_command__, path, body)
            return await execute()
        finally:
            if not provided:
                request_future.cancel()



from whirlwind.commander import Commander, Command
from whirlwind import request_handlers
from whirlwind.version import VERSION
from whirlwind.server import Server
from whirlwind.store import Store

__doc__ = """
A wrapper around the tornado web server.
"""

__all__ = ["VERSION", "Commander", "Command", "Server", "Store", "request_handlers"]



from tornado.httpserver import HTTPServer
import tornado.web
import logging

log = logging.getLogger("whirlwind.server")


class ForcedQuit(Exception):
    pass


class Server(object):
    def __init__(self, final_future, *, server_end_future=None):
        self.final_future = final_future
        if server_end_future is None:
            server_end_future = final_future
        self.server_end_future = server_end_future

    async def serve(self, host, port, *args, **kwargs):
        self.port = port
        self.host = host
        self.server_kwargs = await self.setup(*args, **kwargs)
        if self.server_kwargs is None:
            self.server_kwargs = {}

        self.routes = self.tornado_routes()
        self.http_server = self.make_http_server(self.routes, self.server_kwargs)
        self.announce_start()

        self.http_server.listen(self.port, self.host)
        try:
            await self.wait_for_end()
        except ForcedQuit:
            log.info("The server was told to shut down")
        finally:
            try:
                self.http_server.stop()
            finally:
                await self.cleanup()

    async def wait_for_end(self):
        """Hook that will end when we need to stop the server"""
        await self.server_end_future

    def make_http_server(self, routes, server_kwargs):
        """
        Used to make the http server itself

        We expect it at least has ``listen(port, host)`` and ``stop()``
        """
        return HTTPServer(self.make_application(routes, server_kwargs))

    def make_application(self, routes, server_kwargs):
        """The WSGI application we are starting"""
        return tornado.web.Application(routes, **server_kwargs)

    def announce_start(self):
        """Called after the server has been created and just before it is started"""
        log.info(f"Hosting server at http://{self.host}:{self.port}")

    async def setup(self, *args, **kwargs):
        """
        Hook that receives all extra args and kwargs from serve

        The return of this function is either None or a dictionary of kwargs
        to add to our instantiation of the tornado.web.Application.
        """

    def tornado_routes(self):
        """
        Must be implemented to provide the list of routes given to the tornado.web.Application
        """
        raise NotImplementedError()

    async def cleanup(self):
        """Called after the server has stopped"""



from whirlwind.request_handlers.base import Simple, SimpleWebSocketBase, Finished
from whirlwind.store import NoSuchPath

import logging
import inspect

log = logging.getLogger("whirlwind.request_handlers.command")


class ProgressMessageMaker:
    def __init__(self, stack_level=0):
        mod = None
        try:
            stack = inspect.stack()
            frm = stack[1 + stack_level]
            mod = inspect.getmodule(frm[0])
        except:
            pass

        if mod and hasattr(mod, "__name__"):
            self.logger_name = mod.__name__
        else:
            self.logger_name = "whirlwind.request_handlers.command"

    def __call__(self, body, message, do_log=True, **kwargs):
        info = self.make_info(body, message, **kwargs)
        if do_log:
            self.do_log(body, message, info, **kwargs)
        return info

    def make_info(self, body, message, **kwargs):
        info = {}

        if isinstance(message, Exception):
            info["error_code"] = message.__class__.__name__
            if hasattr(message, "as_dict"):
                info["error"] = message.as_dict()
            else:
                info["error"] = str(message)
        elif message is None:
            info["done"] = True
        elif type(message) is dict:
            info.update(message)
        else:
            info["info"] = message

        info.update(kwargs)
        return info

    def do_log(self, body, message, info, **kwargs):
        pass


class ProcessReplyMixin:
    def process_reply(self, msg, exc_info=None):
        try:
            self.commander.process_reply(msg, exc_info)
        except KeyboardInterrupt:
            raise
        except Exception as error:
            log.exception(error)


class CommandHandler(Simple, ProcessReplyMixin):
    progress_maker = ProgressMessageMaker

    def initialize(self, commander):
        self.commander = commander

    async def do_put(self):
        j = self.body_as_json()

        def progress_cb(message, stack_extra=0, **kwargs):
            maker = self.progress_maker(1 + stack_extra)
            info = maker(j, message, **kwargs)
            self.process_reply(info)

        path = self.request.path
        while path and path.endswith("/"):
            path = path[:-1]

        try:
            return await self.commander.executor(progress_cb, self).execute(path, j)
        except NoSuchPath as error:
            raise Finished(
                status=404,
                wanted=error.wanted,
                available=error.available,
                error="Specified path is invalid",
            )


class WSHandler(SimpleWebSocketBase, ProcessReplyMixin):
    progress_maker = ProgressMessageMaker

    def initialize(self, final_future, server_time, wsconnections, commander):
        self.commander = commander
        super().initialize(final_future, server_time, wsconnections)

    def transform_progress(self, body, progress, stack_extra=0, **kwargs):
        maker = self.progress_maker(2 + stack_extra)
        yield {"progress": maker(body, progress, **kwargs)}

    async def process_message(self, path, body, message_id, message_key, progress_cb):
        try:
            executor = self.commander.executor(
                progress_cb, self, message_key=message_key, message_id=message_id
            )
            return await executor.execute(
                path, body, allow_ws_only=True, request_future=self.connection_future
            )
        except NoSuchPath as error:
            raise Finished(
                status=404,
                wanted=error.wanted,
                available=error.available,
                error="Specified path is invalid",
            )



from whirlwind.commander import Command

from delfick_project.norms import dictobj, sb, BadSpecValue, Meta
from delfick_project.option_merge import NoFormat, MergedOptions
from collections import defaultdict
from textwrap import dedent
import logging
import asyncio
import inspect
import sys


log = logging.getLogger("whirlwind.store")


class CantReuseCommands(Exception):
    def __init__(self, reusing):
        self.reusing = reusing
        super().__init__(
            dedent(
                f"""
            You tried to reuse {self.reusing} as a store command.

            You can use a subclass instead. For example:

                @store.command("command")
                class Example(store.Command):
                    ...

                @store.command("command2")
                class Example2(Example):
                    ...

            Or you can do something like:

                class Base(store.Command):
                    ...

                @store.command("command1")
                class Command1(Base):
                    ...

                @store.command("command2")
                class Command2(Base):
                    ...
        """
            )
        )


def create_task(coro, name=None):
    loop = asyncio.get_event_loop()
    version_info = sys.version_info
    if version_info >= (3, 8):
        return loop.create_task(coro, name=name)
    elif version_info >= (3, 7):
        return loop.create_task(coro)
    else:
        return asyncio.ensure_future(coro)


def retrieve_exception(result):
    if result.cancelled():
        return
    result.exception()


def is_interactive(obj):
    return (
        obj and hasattr(obj, "execute") and "messages" in inspect.signature(obj.execute).parameters
    )


async def pass_on_result(fut, command, execute, *, log_exceptions):
    if execute:
        coro = execute()
    else:
        coro = command.execute()

    def transfer(result):
        if result.cancelled():
            fut.cancel()
            return

        exc = result.exception()
        if fut.done():
            return

        if exc:
            if log_exceptions:
                log.error(exc)
            fut.set_exception(exc)
        else:
            fut.set_result(result.result())

    task = create_task(coro, name=f"<pass_on_result: {command.__class__.__name__}>")
    task.add_done_callback(transfer)
    return await fut


class NoSuchPath(Exception):
    def __init__(self, wanted, available):
        self.wanted = wanted
        self.available = available


class NoSuchParent(Exception):
    def __init__(self, wanted):
        self.wanted = wanted

        s = repr(self.wanted)
        if hasattr(self.wanted, "__name__"):
            s = self.wanted.__name__

        super().__init__(self, f"Couldn't find parent specified by command: {s}")


class NonInteractiveParent(Exception):
    def __init__(self, wanted):
        self.wanted = wanted

        s = repr(self.wanted)
        if hasattr(self.wanted, "__name__"):
            s = self.wanted.__name__

        super().__init__(self, f"Store commands can only specify an interactive parent: {s}")


class ProcessItem:
    def __init__(self, fut, command, execute, messages):
        self.fut = fut
        self.execute = execute
        self.command = command
        self.messages = messages
        self.interactive = is_interactive(self.command)

    def process(self):
        try:
            coro = pass_on_result(
                self.fut, self.command, self.execute, log_exceptions=self.interactive
            )
        except:
            self.fut.set_exception(sys.exc_info()[1])
            return self.fut
        else:
            task = create_task(coro, name=f"<process: {self.command.__class__.__name__}>")
            task.add_done_callback(retrieve_exception)
            self.messages.ts.append((task, False, True))

            return task

    def no_process(self):
        self.fut.set_result({"received": True})
        return self.fut


class MessageHolder:
    def __init__(self, command, final_future):
        self.ts = []
        self.command = command
        self.queue = asyncio.Queue()
        self.final_future = final_future

    def add_main_task(self, main_task):
        self.main_task = main_task

    async def finish(self, cancelled=False):
        exception = None
        if hasattr(self, "main_task") and self.main_task.done():
            if self.main_task.cancelled():
                cancelled = True
            elif self.main_task.exception():
                exception = self.main_task.exception()

        if self.ts:
            for t, do_cancel, do_transfer in self.ts:
                if do_transfer and exception and not t.done():
                    t.set_exception(exception)
                if cancelled or do_cancel:
                    t.cancel()
            await asyncio.wait([t for t, _, _ in self.ts])

    async def add(self, fut, command, execute=None):
        fut.add_done_callback(retrieve_exception)
        self.ts.append((fut, False, True))

        await self.queue.put(ProcessItem(fut, command, execute, self))

    def __aiter__(self):
        return self

    async def __anext__(self):
        while True:
            getter = create_task(
                self.queue.get(), name=f"<queue.get: {self.command.__class__.__name__}>"
            )
            self.ts.append((getter, True, False))
            await asyncio.wait([getter, self.final_future], return_when=asyncio.FIRST_COMPLETED)
            self.ts = [i for i in self.ts if not i[0].done()]

            if self.final_future.done():
                getter.cancel()
                await asyncio.wait([getter])
                raise StopAsyncIteration

            nxt = await getter
            if nxt is not None:
                return nxt


class command_spec(sb.Spec):
    """
    Knows how to turn ``{"path": <string>, "body": {"command": <string>, "args": <dict>}}``
    into the execute method of a Command object.

    It uses the FieldSpec in self.paths to normalise the args into the Command instance.
    """

    def setup(self, paths):
        self.paths = paths
        self.existing_commands = {}

    def make_command(self, meta, val, existing):
        v = sb.set_options(
            path=sb.required(sb.string_spec()), allow_ws_only=sb.defaulted(sb.boolean(), False)
        ).normalise(meta, val)

        path = v["path"]
        allow_ws_only = v["allow_ws_only"]

        if path not in self.paths:
            raise NoSuchPath(path, sorted(self.paths))

        val = sb.set_options(
            body=sb.required(
                sb.set_options(args=sb.dictionary_spec(), command=sb.required(sb.string_spec()))
            )
        ).normalise(meta, val)

        args = val["body"]["args"]
        name = val["body"]["command"]

        if existing:
            name = val["body"]["command"] = f"{existing['path']}:{name}"

        extra_context = {}
        if existing:
            extra_context["_parent_command"] = existing["command"]

        everything = meta.everything
        if isinstance(meta.everything, MergedOptions):
            everything = meta.everything.wrapped()
        everything.update(extra_context)
        meta = Meta(everything, []).at("body")

        available_commands = self.paths[path]

        if name not in available_commands:
            raise BadSpecValue(
                "Unknown command",
                wanted=name,
                available=self.available(available_commands, allow_ws_only=allow_ws_only),
                meta=meta.at("command"),
            )

        command = available_commands[name]["spec"].normalise(meta.at("args"), args)

        if not allow_ws_only and command.__whirlwind_ws_only__:
            raise BadSpecValue(
                "Command is for websockets only",
                wanted=name,
                available=self.available(available_commands, allow_ws_only=allow_ws_only),
                meta=meta.at("command"),
            )

        return command, name

    def available(self, available_commands, *, allow_ws_only):
        available = []
        for name, info in available_commands.items():
            if allow_ws_only or not info["kls"].__whirlwind_ws_only__:
                available.append(name)

        return sorted(available)

    def find_command(self, message_id):
        if isinstance(message_id, tuple) and len(message_id) == 1:
            message_id = message_id[0]

        if not message_id or isinstance(message_id, str):
            return None, (message_id,)

        parent = message_id[:-1]
        if parent not in self.existing_commands:
            raise NoSuchParent(wanted=parent)

        return self.existing_commands[parent], message_id

    def normalise_filled(self, meta, val):
        parent_existing, message_id_tuple = self.find_command(meta.everything.get("message_id"))
        command, path = self.make_command(meta, val, parent_existing)

        existing = None
        if command and is_interactive(command):
            existing = {"command": command, "messages": None, "path": path}
            self.existing_commands[message_id_tuple] = existing

        async def execute():
            if parent_existing and not existing:
                fut = asyncio.Future()
                await parent_existing["messages"].add(fut, command, ())
                return await fut

            try:
                if not existing:
                    return await command.execute()
                else:
                    request_future = meta.everything.get("request_future")
                    return await self.execute_interactive(
                        request_future, parent_existing, existing, command
                    )
            finally:
                if message_id_tuple in self.existing_commands:
                    del self.existing_commands[message_id_tuple]

        execute.__whirlwind_command__ = command
        return execute

    async def execute_interactive(self, request_future, parent_existing, existing, command):
        holder_kls = MessageHolder
        if hasattr(command, "MessageHolder"):
            holder_kls = command.MessageHolder
        existing["messages"] = holder_kls(command, request_future)

        async def execute():
            try:
                task = create_task(
                    existing["command"].execute(existing["messages"]),
                    name=f"<execute interactive: {command.__class__.__name__}",
                )
                existing["messages"].add_main_task(task)
                return await task
            finally:
                cancelled = False
                exc_info = sys.exc_info()
                if exc_info and isinstance(exc_info[1], asyncio.CancelledError):
                    cancelled = True

                if isinstance(existing, dict) and existing["messages"]:
                    try:
                        await existing["messages"].finish(cancelled=cancelled)
                    except asyncio.CancelledError:
                        raise
                    except:
                        log.exception("Failed to finish the message holder")

        if parent_existing:
            fut = asyncio.Future()
            await parent_existing["messages"].add(fut, command, execute)
            return await fut
        else:
            return await execute()


class Store:
    Command = Command

    _merged_options_formattable = True

    def __init__(self, prefix=None, default_path="/v1", formatter=None):
        self.prefix = self.normalise_prefix(prefix)
        self.formatter = formatter
        self.default_path = default_path
        self.paths = defaultdict(dict)
        self.command_spec = command_spec(self.paths)

    def clone(self):
        new_store = Store(self.prefix, self.default_path, self.formatter)
        for path, commands in self.paths.items():
            new_store.paths[path].update(dict(commands))
        return new_store

    def injected(self, path, format_into=sb.NotSpecified, nullable=False):
        class find_value(sb.Spec):
            def normalise(s, meta, val):
                if nullable and path not in meta.everything:
                    return NoFormat(None)
                return f"{{{path}}}"

        return dictobj.Field(find_value(), formatted=True, format_into=format_into)

    def normalise_prefix(self, prefix, trailing_slash=True):
        if prefix is None:
            return ""

        if trailing_slash:
            if prefix and not prefix.endswith("/"):
                return f"{prefix}/"
        else:
            while prefix and prefix.endswith("/"):
                prefix = prefix[:-1]

        return prefix

    def normalise_path(self, path):
        if path is None:
            path = self.default_path

        while path and path.endswith("/"):
            path = path[:-1]

        if not path.startswith("/"):
            path = f"/{path}"

        return path

    def merge(self, other, prefix=None):
        new_prefix = self.normalise_prefix(prefix, trailing_slash=False)
        for path, commands in other.paths.items():
            for name, options in commands.items():
                slash = ""
                if not new_prefix.endswith("/") and not name.startswith("/"):
                    slash = "/"
                self.paths[path][f"{new_prefix}{slash}{name}"] = options

    def command(self, name, *, path=None, parent=None):
        path = self.normalise_path(path)

        def decorator(kls):
            if "__whirlwind_command__" in kls.__dict__:
                raise CantReuseCommands(kls)

            kls.__whirlwind_command__ = True
            kls.__whirlwind_ws_only__ = is_interactive(kls) or parent

            n = name
            spec = kls.FieldSpec(formatter=self.formatter)

            if parent and not is_interactive(parent):
                raise NonInteractiveParent(parent)
            elif parent:
                found = False
                for p, o in self.paths[path].items():
                    if o["kls"] is parent:
                        n = f"{p}:{name}"
                        found = True
                        break

                if not found:
                    raise NoSuchParent(parent)
            else:
                n = f"{self.prefix}{n}"

            self.paths[path][n] = {"kls": kls, "spec": spec}
            return kls

        return decorator



from whirlwind.request_handlers.base import Finished, MessageFromExc, Simple, SimpleWebSocketBase
from whirlwind.request_handlers.command import ProgressMessageMaker, CommandHandler, WSHandler

__all__ = [
    "Finished",
    "MessageFromExc",
    "Simple",
    "SimpleWebSocketBase",
    "ProgressMessageMaker",
    "CommandHandler",
    "WSHandler",
]



from whirlwind.store import create_task

from delfick_project.norms import sb, dictobj, Meta
from tornado.web import RequestHandler, HTTPError
from tornado import websocket
import binascii
import logging
import asyncio
import json
import uuid

log = logging.getLogger("whirlwind.request_handlers.base")


class Finished(Exception):
    def __init__(self, status=500, **kwargs):
        self.kwargs = kwargs
        self.kwargs["status"] = status

    def as_dict(self):
        return self.kwargs


def reprer(o):
    if type(o) is bytes:
        return binascii.hexlify(o).decode()
    return repr(o)


class MessageFromExc:
    def __init__(self, *, log_exceptions=True, see_exception=None):
        self.see_exception = see_exception
        self.log_exceptions = log_exceptions

    def __call__(self, exc_type, exc, tb):
        if isinstance(exc, Finished):
            return exc.kwargs
        else:
            return self.process(exc_type, exc, tb)

    def process(self, exc_type, exc, tb):
        if self.see_exception:
            self.see_exception(exc_type, exc, tb)

        if exc_type is asyncio.CancelledError:
            return {
                "status": 500,
                "error": "Request was cancelled",
                "error_code": "RequestCancelled",
            }
        else:
            if self.see_exception is None and self.log_exceptions:
                log.error(exc, exc_info=(exc_type, exc, tb))

            return {
                "status": 500,
                "error": "Internal Server Error",
                "error_code": "InternalServerError",
            }


class AsyncCatcher(object):
    def __init__(self, request, info, final=None):
        self.info = info
        self.final = final
        self.request = request

    async def __aenter__(self):
        pass

    async def __aexit__(self, exc_type, exc, tb):
        if exc is None:
            self.complete(self.info.get("result"), status=200)
            return

        msg = self.request.message_from_exc(exc_type, exc, tb)
        self.complete(msg, status=500, exc_info=(exc_type, exc, tb))

        # And don't reraise the exception
        return True

    def send_msg(self, msg, status=200, exc_info=None):
        if self.request._finished and not hasattr(self.request, "ws_connection"):
            if type(msg) is dict:
                msg = json.dumps(msg, default=self.request.reprer, sort_keys=True, indent="    ")
                self.request.hook("request_already_finished", msg)
            return

        if hasattr(msg, "exc_info") and exc_info is None:
            exc_info = msg.exc_info

        if self.final is None:
            self.request.send_msg(msg, status, exc_info=exc_info)
        else:
            self.final(msg, exc_info=exc_info)

    def complete(self, msg, status=sb.NotSpecified, exc_info=None):
        if type(msg) is dict:
            result = json.loads(json.dumps(msg, default=self.request.reprer, indent="    "))
        else:
            result = msg

        self.send_msg(result, status=status, exc_info=exc_info)


class RequestsMixin:
    """
    A mixin class you may use for your handler which provides some handy methods
    for dealing with data
    """

    _merged_options_formattable = True

    def hook(self, func, *args, **kwargs):
        if hasattr(self, func):
            return getattr(self, func)(*args, **kwargs)

    # def process_reply(self, msg, exc_info=None):
    #     """A hook that provides the msg sent as reply or progress"""
    #     pass

    # def request_already_finished(self, msg):
    #     """Hook for when we would send a message to an already closed websocket"""

    @property
    def reprer(self):
        if not hasattr(self, "_reprer"):
            self._reprer = reprer
        return self._reprer

    @reprer.setter
    def reprer(self, value):
        self._reprer = value

    @property
    def message_from_exc(self):
        if not hasattr(self, "_message_from_exc"):
            self._message_from_exc = MessageFromExc(
                see_exception=self.see_returned_exception,
                log_exceptions=getattr(self, "log_exceptions", True),
            )
        return self._message_from_exc

    def see_returned_exception(self, exc_typ, exc, tb):
        if getattr(self, "log_exceptions", True):
            log.error(exc, exc_info=(exc_typ, exc, tb))

    @message_from_exc.setter
    def message_from_exc(self, value):
        self._message_from_exc = value

    def async_catcher(self, info, final=None):
        return AsyncCatcher(self, info, final=final)

    def body_as_json(self, body=None):
        """
        Return the body of the request as a json object

        If there is a special ``__body__`` file in the request, we will consider this
        to be the body instead of the request body
        """
        if body is None:
            if "__body__" in self.request.files:
                body = self.request.files["__body__"][0]["body"].decode()
            else:
                body = self.request.body.decode()

        try:
            if type(body) is str:
                body = json.loads(body)
        except (TypeError, ValueError) as error:
            self.log_json_error(body, error)
            raise Finished(status=400, reason="Failed to load body as json", error=error)

        return body

    def log_json_error(self, body, error):
        """
        Do any logging for errors parsing the request body as json
        """
        log.error("Failed to load body as json\t%s", body)

    def send_msg(self, msg, status=sb.NotSpecified, exc_info=None):
        """
        This determines what content-type and exact body to write to the response

        If ``msg`` has ``as_dict``, we call it.

        If ``msg`` is a dictionary and has status, we use that as the status of
        the request, otherwise we say it's a 200.

        If there is ``html`` in ``msg``, we use that as the body of the request.

        If ``msg`` is None, we close without a body.

        * If ``msg`` is a ``dict`` or ``list``, we write it as a json object.
        * If ``msg`` starts with ``<html>`` or ``<!DOCTYPE html>`` we treat it
          as html content
        * Otherwise we write ``msg`` as ``text/plain``
        """
        if hasattr(msg, "exc_info") and exc_info is None:
            exc_info = msg.exc_info

        if hasattr(msg, "as_dict"):
            msg = msg.as_dict()

        self.hook("process_reply", msg, exc_info=exc_info)

        if type(msg) is dict and "status" in msg:
            status = msg["status"]
        elif exc_info and exc_info[1]:
            if hasattr(exc_info[1], "status"):
                status = exc_info[1].status
            else:
                status = 500

        if status is sb.NotSpecified:
            status = 200

        self.set_status(status)

        if type(msg) is dict and "html" in msg:
            msg = msg["html"]

        if msg is None:
            self.finish()
            return

        if type(msg) in (dict, list):
            self.set_header("Content-Type", "application/json; charset=UTF-8")
            self.write(json.dumps(msg, default=self.reprer, sort_keys=True, indent="    "))
        elif msg.lstrip().startswith("<html>") or msg.lstrip().startswith("<!DOCTYPE html>"):
            self.write(msg)
        else:
            self.set_header("Content-Type", "text/plain; charset=UTF-8")
            self.write(msg)
        self.finish()


class Simple(RequestsMixin, RequestHandler):
    """
    Helper for using ``self.async_catcher`` from ``RequestsMixin`` for most HTTP verbs.

    .. code-block:: python

        class MyRequestHandler(Simple):
            async def do_get():
                return "<html><body><p>lol</p></body></html>"

    Essentially you define ``async def do_<verb>(self)`` methods for each verb
    you want to support.

    This supports

    * get
    * put
    * post
    * patch
    * delete
    """

    log_exceptions = True

    async def get(self, *args, **kwargs):
        if not hasattr(self, "do_get"):
            raise HTTPError(405)

        info = {"result": None}
        async with self.async_catcher(info):
            info["result"] = await self.do_get(*args, **kwargs)

    async def put(self, *args, **kwargs):
        if not hasattr(self, "do_put"):
            raise HTTPError(405)

        info = {"result": None}
        async with self.async_catcher(info):
            info["result"] = await self.do_put(*args, **kwargs)

    async def post(self, *args, **kwargs):
        if not hasattr(self, "do_post"):
            raise HTTPError(405)

        info = {"result": None}
        async with self.async_catcher(info):
            info["result"] = await self.do_post(*args, **kwargs)

    async def patch(self, *args, **kwargs):
        if not hasattr(self, "do_patch"):
            raise HTTPError(405)

        info = {"result": None}
        async with self.async_catcher(info):
            info["result"] = await self.do_patch(*args, **kwargs)

    async def delete(self, *args, **kwargs):
        if not hasattr(self, "do_delete"):
            raise HTTPError(405)

        info = {"result": None}
        async with self.async_catcher(info):
            info["result"] = await self.do_delete(*args, **kwargs)


json_spec = sb.match_spec(
    (bool, sb.any_spec()),
    (int, sb.any_spec()),
    (float, sb.any_spec()),
    (str, sb.any_spec()),
    (list, lambda: sb.listof(json_spec)),
    (type(None), sb.any_spec()),
    fallback=lambda: sb.dictof(sb.string_spec(), json_spec),
)


class SimpleWebSocketBase(RequestsMixin, websocket.WebSocketHandler):
    """
    Used for websocket handlers

    Implement ``process_message``

    .. automethod:: whirlwind.request_handlers.base.SimpleWebSocketBase.process_message

    This class takes in messages of the form ``{"path": <string>, "message_id": <string>, "body": <dictionary}``

    It will respond with messages of the form ``{"reply": <reply>, "message_id": <message_id>}``

    It treats path of ``__tick__`` as special and respond with ``{"reply": {"ok": "thankyou"}, "message_id": "__tick__"}``

    It relies on the client side closing the connection when it's finished.
    """

    log_exceptions = True

    def initialize(self, final_future, server_time, wsconnections):
        self.server_time = server_time
        self.final_future = final_future
        self.wsconnections = wsconnections

    class WSMessage(dictobj.Spec):
        path = dictobj.Field(sb.string_spec, wrapper=sb.required)
        message_id = dictobj.Field(
            sb.or_spec(sb.string_spec(), sb.tupleof(sb.string_spec())), wrapper=sb.required
        )
        body = dictobj.Field(json_spec, wrapper=sb.required)

    message_spec = WSMessage.FieldSpec()

    class Closing(object):
        pass

    def open(self):
        self.key = str(uuid.uuid1())
        self.connection_future = asyncio.Future()
        if self.final_future.done():
            self.connection_future.cancel()
            return

        canceller = lambda res: self.connection_future.cancel()
        self.final_future.add_done_callback(canceller)
        self.connection_future.add_done_callback(
            lambda res: self.final_future.remove_done_callback(canceller)
        )

        if self.server_time is not None:
            self.reply(self.server_time, message_id="__server_time__")
        self.hook("websocket_opened")

    def reply(self, msg, message_id=None, exc_info=None):
        if msg is None:
            msg = {"done": True}

        # I bypass tornado converting the dictionary so that non jsonable things can be repr'd
        if hasattr(msg, "as_dict"):
            msg = msg.as_dict()
        reply = {"reply": msg, "message_id": message_id}
        reply = json.dumps(reply, default=self.reprer).replace("</", "<\\/")

        if message_id not in ("__tick__", "__server_time__"):
            self.hook("process_reply", msg, exc_info=exc_info)

        if self.ws_connection:
            self.write_message(reply)

    def on_message(self, message):
        self.hook("websocket_message", message)
        try:
            parsed = json.loads(message)
        except (TypeError, ValueError) as error:
            self.reply({"error": "Message wasn't valid json\t{0}".format(str(error))})
            return

        if type(parsed) is dict and "path" in parsed and parsed["path"] == "__tick__":
            parsed["message_id"] = "__tick__"
            parsed["body"] = "__tick__"

        try:
            msg = self.message_spec.normalise(Meta.empty(), parsed)
        except Exception as error:
            self.hook("websocket_invalid_message", error, parsed)

            if hasattr(error, "as_dict"):
                error = error.as_dict()
            else:
                error = str(error)

            self.reply({"error_code": "InvalidMessage", "error": error})
        else:
            path = msg.path
            body = msg.body
            message_id = msg.message_id
            message_key = str(uuid.uuid4())

            if path == "__tick__":
                self.reply({"ok": "thankyou"}, message_id=message_id)
                return

            def on_processed(final, exc_info=None):
                if final is self.Closing:
                    self.reply({"closing": "goodbye"}, message_id=message_id)
                    self.close()
                else:
                    self.reply(final, message_id=message_id, exc_info=exc_info)

                try:
                    self.message_done(msg, final, message_key, exc_info=exc_info)
                except Exception as error:
                    self.handle_message_done_error(error, msg, final, message_key, exc_info)

            async def doit():
                info = {}

                def progress_cb(progress, **kwargs):
                    for m in self.transform_progress(msg, progress, **kwargs):
                        self.reply(m, message_id=message_id)

                async with self.async_catcher(info, on_processed):
                    result = await self.process_message(
                        path, body, message_id, message_key, progress_cb
                    )

                    if isinstance(result, asyncio.Future) or hasattr(result, "__await__"):
                        result = await result

                    info["result"] = result

            def done(res):
                if message_key in self.wsconnections:
                    del self.wsconnections[message_key]

                if not res.cancelled():
                    self.handle_request_done_exception(res.exception())

            t = create_task(doit(), name=f"<process_command: {body}>")
            t.add_done_callback(done)
            self.wsconnections[message_key] = t

    def message_done(self, request, final, message_key, exc_info=None):
        """
        Hook for when we have finished processing a request

        By default nothing is done.

        request
            The original request

        final
            The last response to be sent back.

        message_key
            The uuid the server generated for this request

        exc_info
            The (exc_type, exc, traceback) for any exception that stopped the processing of the request
        """

    def handle_message_done_error(self, error, msg, final, message_key, exc_info):
        """
        Hook for when message_done raised an exception

        By default we ``log.exception(error)``

        error
            The exception that was raised

        request
            The original request

        final
            The last response to be sent back.

        message_key
            The uuid the server generated for this request

        exc_info
            The (exc_type, exc, traceback) for any exception that stopped the processing of the request
            before message_done had been called
        """
        log.exception(error)

    def handle_request_done_exception(self, error):
        """Hook that takes in an exception from the entire request"""
        if error and self.log_exceptions:
            log.exception(error, exc_info=(type(error), error, error.__traceback__))

    def transform_progress(self, body, progress, **kwargs):
        """
        Hook for transforming progress messages. This must be a generator that yields 0 or more messages

        So when the ``progress_cb`` is called like ``progress_cb("some message", arg=1)`` we will do:

        .. code-block:: python

            for m in self.transform_progress(<request>, "some message", arg=1):
                # write ``{"reply": m, "message_id": <message_id>}``

        where ``<request>`` is the entire message that started this stream.

        By default kwargs are ignored and we just yield ``{"progress": progress}`` once
        """
        yield {"progress": progress}

    async def process_message(self, path, body, message_id, message_key, progress_cb):
        """
        Return the response to be sent back when we get a message from the conn.

        path
            The uri specified in the message

        body
            The body specified in the message

        message_id
            The unique message_id for this stream of requests as supplied in the request

        message_key
            A unique id for this stream created by the server

        progress_cb
            A callback that will send a message of the form ``{"progress": <progress>, "message_id": <message_id}``
            where ``<progress>`` is the argument passed into the callback
        """
        raise NotImplementedError

    def on_close(self):
        """Hook for when a websocket connection closes"""
        self.connection_future.cancel()



from setuptools import setup, find_packages
import runpy
import os

version = runpy.run_path(os.path.join(os.path.dirname(__file__), "whirlwind", "version.py"))

# fmt: off

setup(
      name = "whirlwind-web"
    , version = version["VERSION"]
    , packages = find_packages(include="whirlwind.*", exclude=["tests*"])

    , install_requires =
      [ "tornado >= 5.1.1"
      , "delfick_project >= 0.5"
      ]

    , extras_require =
      { "tests":
        [ "noseOfYeti==2.0.2"
        , "mock==4.0.2"
        , "pytest==5.3.1"
        , "aiohttp==3.7.0"
        , "alt-pytest-asyncio==0.5.3"
        , "pytest-helpers-namespace==2019.1.8"
        ]
      , "peer":
        [ "tornado==5.1.1"
        , "delfick_project==0.5"
        ]
      }

    , entry_points =
      { 'console_scripts' :
        [ 'run_whirlwind_pytest = whirlwind._pytest:run_pytest'
        ]
      }

    # metadata for upload to PyPI
    , url = "http://github.com/delfick/whirlwind"
    , author = "Stephen Moore"
    , author_email = "delfick755@gmail.com"
    , description = "Wrapper around the tornado web server library"
    , long_description = open("README.rst").read()
    , license = "MIT"
    , keywords = "tornado web"
    )

# fmt: on



import pytest
import sys


def run_pytest():
    class EditConfig:
        @pytest.hookimpl(hookwrapper=True)
        def pytest_cmdline_parse(pluginmanager, args):
            args.extend(
                [
                    "--tb=short",
                    "-o",
                    "console_output_style=classic",
                    "-o",
                    "default_alt_async_timeout=1",
                    "-p",
                    "helpers_namespace",
                ]
            )
            yield

    sys.exit(pytest.main(plugins=[EditConfig()]))
